
#     ________
#            /
#      \    /
#       \  /
#        \/

# Main reference: Qian (2005)
# E. Qian. 2005. Risk Parity Portfolios: Efficient Portfolios Through True Diversification.
# White paper. Panagora Asset Management, Boston, Massachusetts, USA.

import numpy as np
from scipy.optimize import minimize


def cons_sum_weight(x):
    """
    Constraint function - weights must sum to one.

    Parameters
    ----------
    x : real ndarray
        Weights of assets in portfolio.

    Returns
    -------
    weights_sum_zero : float
        Weights summed and one subtracted from them - constraint must be zero.

    Notes
    -----
    Equality constraint: = 0

    """
    weights_sum_zero = np.sum(x) - 1

    return weights_sum_zero


def cons_long_only_weight(x):
    """
    Constraint function - weights must all be non-negative.

    Parameters
    ----------
    x : real ndarray
        Weights of assets in portfolio.

    Returns
    -------
    x : real ndarray
        Weights must all be non-negative.

    Notes
    -----
    Inequality constraint: > 0

    """
    return x


def risk_parity_obj_fun(x, p_cov, rb):
    """
    Risk Parity or Risk Premia Parity objective function.

    Parameters
    ----------
    x : real ndarray
        Weights of assets in portfolio.

    p_cov : real ndarray
        Covariance matrix of portfolio.

    rb : real ndarray
        The target weighting vector. i.e. if [1/n, ..., 1/n] then risk is equally allocated.

    Returns
    -------
    risk_budget_obj : float
        Risk budgeting objective function value.

    Notes
    -----

    """

    if not isinstance(x, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'x\' must be of type np.ndarray.')
    if np.array(x).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'x\' must only contain floats.')
    if not isinstance(p_cov, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'p_cov\' must be of type np.ndarray.')
    if np.array(p_cov).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'p_cov\' must only contain floats.')
    if not np.all(np.linalg.eigvals(p_cov) > 0):
        raise ValueError('\'p_cov\' must be PSD.')
    if len(x) != np.shape(p_cov)[0] or len(x) != np.shape(p_cov)[1]:
        raise ValueError('\'x\' and \'p_cov\' are incompatible lengths.')
    if not isinstance(rb, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'rb\' must be of type np.ndarray.')
    if np.array(rb).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'rb\' must only contain floats.')
    if len(x) != len(rb):
        raise ValueError('\'x\' and \'rb\' are incompatible lengths.')

    risk_budget_obj = np.sum((x * np.dot(p_cov, x) / np.dot(x.transpose(), np.dot(p_cov, x)) - rb) ** 2)

    return risk_budget_obj


def equal_risk_parity_weights_long_restriction(cov):
    """
    Risk Parity or Risk Premia Parity minimizer with long restriction.

    Parameters
    ----------
    cov : real ndarray
        Covariance matrix of portfolio.

    Returns
    -------
    OptimizeResult : OptimizeResult
        Optimised result and optimised weights.

    Notes
    -----
    Risk Premia Parity weighting with long restriction.
    Constrained optimisation - 'SLSQP' won't change if variance is too low - must change 'ftol' to smaller value.

    """

    if not isinstance(cov, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'cov\' must be of type np.ndarray.')
    if np.array(cov).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'cov\' must only contain floats.')
    if not np.all(np.linalg.eigvals(cov) > 0):
        raise ValueError('\'cov\' must be PSD.')

    w0 = np.ones(np.shape(cov)[0]) / np.shape(cov)[0]
    cons = ({'type': 'eq', 'fun': cons_sum_weight},
            {'type': 'ineq', 'fun': cons_long_only_weight})
    return minimize(risk_parity_obj_fun, w0, args=(cov, (1 / np.shape(cov)[0]) * np.ones_like(w0)),
                    method='SLSQP', constraints=cons, options={'ftol': 1e-9})


def cons_short_limit_weight(x, k):
    """
    Constraint function - weights must all be individually greater than -k.

    Parameters
    ----------
    x : real ndarray
        Weights of assets in portfolio.

    k : positive float
        Individual weights must each be greater than -k.

    Returns
    -------
    y : real ndarray
        Weights must all be greater than -k.

    Notes
    -----
    Inequality constraint: > -k.
    Require a summation restriction as with a large number of assets this can grow unreasonably large.

    """
    y = x + k

    return y


def cons_long_limit_weight(x, k):
    """
    Constraint function - weights must all be individually less than k.

    Parameters
    ----------
    x : real ndarray
        Weights of assets in portfolio.

    k : positive float
        Individual weights must each be less than k.

    Returns
    -------
    y : real ndarray
        Weights must all be less than k.

    Notes
    -----
    Inequality constraint: < k.
    Require a summation restriction as with a large number of assets this can grow unreasonably large.

    """
    y = x - k

    return y


def equal_risk_parity_weights_short_restriction(cov, short_limit=1.0):
    """
    Risk Parity or Risk Premia Parity minimizer with individual shorting restriction.

    Parameters
    ----------
    cov : real ndarray
        Covariance matrix of portfolio.

    short_limit : positive float
        Individual asset shorting restriction.

    Returns
    -------
    OptimizeResult : OptimizeResult
        Optimised result and optimised weights.

    Notes
    -----
    Risk Premia Parity weighting with individual shorting restriction.
    Constrained optimisation - 'SLSQP' won't change if variance is too low - must change 'ftol' to smaller value.

    """

    if not isinstance(cov, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'cov\' must be of type np.ndarray.')
    if np.array(cov).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'cov\' must only contain floats.')
    if not np.all(np.linalg.eigvals(cov) > 0):
        raise ValueError('\'cov\' must be PSD.')
    if (not isinstance(short_limit, float)) or short_limit <= 0:
        raise ValueError('\'short_limit\' must be a positive float.')

    w0 = np.ones(np.shape(cov)[0]) / np.shape(cov)[0]
    cons = ({'type': 'eq', 'fun': cons_sum_weight},
            {'type': 'ineq', 'fun': cons_short_limit_weight, 'args': [short_limit]})
    return minimize(risk_parity_obj_fun, w0, args=(cov, (1 / np.shape(cov)[0]) * np.ones_like(w0)),
                    method='SLSQP', constraints=cons, options={'ftol': 1e-9})


def cons_short_limit_sum_weight(x, k):
    """
    Constraint function - combined shorting weights must be greater than -k.

    Parameters
    ----------
    x : real ndarray
        Weights of assets in portfolio.

    k : positive float
        Summed negative weights must be greater than -k.

    Returns
    -------
    y : real ndarray
        Summed negative weights must be greater than -k.

    Notes
    -----
    Inequality constraint: > -k.

    """
    y = np.sum(x[x < 0]) + k

    return y


def cons_long_limit_sum_weight(x, k):
    """
    Constraint function - combined long weights must be less than k.

    Parameters
    ----------
    x : real ndarray
        Weights of assets in portfolio.

    k : positive float
        Summed positive weights must be lesser than k.

    Returns
    -------
    y : real ndarray
        Summed positive weights must be less than k.

    Notes
    -----
    Inequality constraint: < k.

    """
    y = np.sum(x[x > 0]) - k

    return y


def equal_risk_parity_weights_summation_restriction(cov, short_limit=0.3, long_limit=1.3):
    """
    Risk Parity or Risk Premia Parity minimizer with shorting summation restriction.
    Base example is the classic 130/30 strategy or 130/30 Long/Short Equity example.

    Parameters
    ----------
    cov : real ndarray
        Covariance matrix of portfolio.

    short_limit : positive float
        Negative weights summation asset shorting restriction.

    long_limit : positive float
        Positive weights summation asset long restriction.

    Returns
    -------
    OptimizeResult : OptimizeResult
        Optimised result and optimised weights.

    Notes
    -----
    Risk Premia Parity weighting with long and short summation restriction.
    Constrained optimisation - 'SLSQP' won't change if variance is too low - must change 'ftol' to smaller value.

    """

    if not isinstance(cov, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'cov\' must be of type np.ndarray.')
    if np.array(cov).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'cov\' must only contain floats.')
    if not np.all(np.linalg.eigvals(cov) > 0):
        raise ValueError('\'cov\' must be PSD.')
    if (not isinstance(short_limit, float)) or short_limit <= 0:
        raise ValueError('\'short_limit\' must be a positive float.')
    if (not isinstance(long_limit, float)) or long_limit <= 0:
        raise ValueError('\'long_limit\' must be a positive float.')

    w0 = np.ones(np.shape(cov)[0]) / np.shape(cov)[0]
    cons = ({'type': 'eq', 'fun': cons_sum_weight},
            {'type': 'ineq', 'fun': cons_short_limit_sum_weight, 'args': [short_limit]},
            {'type': 'ineq', 'fun': cons_short_limit_sum_weight, 'args': [long_limit]})
    return minimize(risk_parity_obj_fun, w0, args=(cov, (1 / np.shape(cov)[0]) * np.ones_like(w0)),
                    method='SLSQP', constraints=cons, options={'ftol': 1e-9})


# The below functions are for the global minimum variance portfolios.


def global_obj_fun(x, p_cov):
    """
    Global minimum variance objective function.

    Parameters
    ----------
    x : real ndarray
        Weights of assets in portfolio.

    p_cov : real ndarray
        Covariance matrix of portfolio.

    Returns
    -------
    obj : float
        Global minimum variance objective function value.

    Notes
    -----

    """

    if not isinstance(x, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'x\' must be of type np.ndarray.')
    if np.array(x).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'x\' must only contain floats.')
    if not isinstance(p_cov, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'p_cov\' must be of type np.ndarray.')
    if np.array(p_cov).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'p_cov\' must only contain floats.')
    if not np.all(np.linalg.eigvals(p_cov) > 0):
        raise ValueError('\'p_cov\' must be PSD.')
    if len(x) != np.shape(p_cov)[0] or len(x) != np.shape(p_cov)[1]:
        raise ValueError('\'x\' and \'p_cov\' are incompatible lengths.')

    obj = np.sum(x * np.dot(p_cov, x))

    return obj


def global_weights(cov):
    """
    Global minimum variance weights direct calculation.

    Parameters
    ----------
    cov : real ndarray
        Covariance matrix of portfolio.

    Returns
    -------
    weights : real ndarray
        Global minimum variance weights.

    Notes
    -----

    """

    if not isinstance(cov, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'cov\' must be of type np.ndarray.')
    if np.array(cov).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'cov\' must only contain floats.')
    if not np.all(np.linalg.eigvals(cov) > 0):
        raise ValueError('\'cov\' must be PSD.')

    try:
        weights = (np.matmul(np.linalg.inv(cov), np.ones(np.shape(cov)[1]).reshape(-1, 1)) /
                   np.matmul(np.ones(np.shape(cov)[1]).reshape(1, -1),
                             np.matmul(np.linalg.inv(cov), np.ones(np.shape(cov)[1]).flatten())))[:, 0]

        return weights
    except:
        weights = (np.matmul(np.linalg.pinv(cov), np.ones(np.shape(cov)[1]).reshape(-1, 1)) /
                   np.matmul(np.ones(np.shape(cov)[1]).reshape(1, -1),
                             np.matmul(np.linalg.pinv(cov), np.ones(np.shape(cov)[1]).flatten())))[:, 0]

        return weights


def global_weights_long(cov):
    """
    Global minimum variance weights constrained calculation.

    Parameters
    ----------
    cov : real ndarray
        Covariance matrix of portfolio.

    Returns
    -------
    OptimizeResult : OptimizeResult
        Optimised result and optimised weights.

    Notes
    -----
    Global minimum variance portfolio weighting with long restriction.
    Constrained optimisation - 'SLSQP' won't change if variance is too low - must change 'ftol' to smaller value.

    """

    if not isinstance(cov, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'cov\' must be of type np.ndarray.')
    if np.array(cov).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'cov\' must only contain floats.')
    if not np.all(np.linalg.eigvals(cov) > 0):
        raise ValueError('\'cov\' must be PSD.')

    w0 = np.ones(np.shape(cov)[0]) / np.shape(cov)[0]
    cons = ({'type': 'eq', 'fun': cons_sum_weight},
            {'type': 'ineq', 'fun': cons_long_only_weight})
    return minimize(global_obj_fun, w0, args=(cov),
                    method='SLSQP', constraints=cons, options={'ftol': 1e-9})


def global_weights_short_and_long_restrict(cov, a, b):
    """
    Global minimum variance weight calculation with constraints on individual weight longing and shorting..

    Parameters
    ----------
    cov : real ndarray
        Covariance matrix of portfolio.

    a : float
        Individual long limit.

    b : float
        Individual short limit.

    Returns
    -------
    OptimizeResult : OptimizeResult
        Optimised result and optimised weights.

    Notes
    -----
    Global minimum variance portfolio weighting with long and short restriction on individual weights.
    Constrained optimisation - 'SLSQP' won't change if variance is too low - must change 'ftol' to smaller value.

    Recommended above direct calculation via inverse and pseudo inverse as inverting large matrices is unstable.

    """

    if not isinstance(cov, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'cov\' must be of type np.ndarray.')
    if np.array(cov).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'cov\' must only contain floats.')
    if not np.all(np.linalg.eigvals(cov) > 0):
        raise ValueError('\'cov\' must be PSD.')
    if (not isinstance(b, float)) or b <= 0:
        raise ValueError('\'b\' must be a positive float.')
    if (not isinstance(a, float)) or a <= 0:
        raise ValueError('\'a\' must be a positive float.')

    w0 = np.ones(np.shape(cov)[0]) / np.shape(cov)[0]
    cons = ({'type': 'eq', 'fun': cons_sum_weight},
            {'type': 'ineq', 'fun': cons_long_limit_weight, 'args': [a]},
            {'type': 'ineq', 'fun': cons_short_limit_weight, 'args': [b]})
    return minimize(global_obj_fun, w0, args=(cov),method='SLSQP', constraints=cons, options={'ftol': 1e-30})


def equal_risk_parity_weights_individual_restriction(cov, short_limit=1.0, long_limit=1.0):
    """
    Risk Parity or Risk Premia Parity minimizer with long and shorting summation restriction.
    Function created and designed to prevent exceedingly large weights in optimisation pursuit.

    Parameters
    ----------
    cov : real ndarray
        Covariance matrix of portfolio.

    short_limit : positive float
        Negative weights summation asset shorting restriction.

    long_limit : positive float
        Positive weights summation asset long restriction.

    Returns
    -------
    OptimizeResult : OptimizeResult
        Optimised result and optimised weights.

    Notes
    -----
    Risk Premia Parity weighting with long and short summation restriction.
    Constrained optimisation - 'SLSQP' won't change if variance is too low - must change 'ftol' to smaller value.

    """

    if not isinstance(cov, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'cov\' must be of type np.ndarray.')
    if np.array(cov).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'cov\' must only contain floats.')
    if not np.all(np.linalg.eigvals(cov) > 0):
        raise ValueError('\'cov\' must be PSD.')
    if (not isinstance(short_limit, float)) or short_limit <= 0:
        raise ValueError('\'short_limit\' must be a positive float.')
    if (not isinstance(long_limit, float)) or long_limit <= 0:
        raise ValueError('\'long_limit\' must be a positive float.')

    w0 = np.ones(np.shape(cov)[0]) / np.shape(cov)[0]
    cons = ({'type': 'eq', 'fun': cons_sum_weight},
            {'type': 'ineq', 'fun': cons_long_limit_weight, 'args': [long_limit]},
            {'type': 'ineq', 'fun': cons_short_limit_weight, 'args': [short_limit]})
    return minimize(risk_parity_obj_fun, w0, args=(cov, (1 / np.shape(cov)[0]) * np.ones_like(w0)),
                    method='SLSQP', constraints=cons, options={'ftol': 1e-9})




#     ________
#            /
#      \    /
#       \  /
#        \/

# Main reference: Hoff and Niu (2012)
# Hoff, P. and Niu, X., A Covariance Regression Model.
# Statistica Sinica, Institute of Statistical Science, 2012, 22(2), 729–753.

import numpy as np
import group_lasso
from sklearn import linear_model

np.random.seed(0)


def calc_B_Psi(m: np.ndarray, v: np.ndarray, x: np.ndarray, y: np.ndarray, basis: np.ndarray, A_est: np.ndarray,
               technique: str, alpha: float, l1_ratio_or_reg: float, group_reg: float, max_iter: int,
               groups: np.ndarray):
    """
    This follows the calculation at the bottom of page 10 and top of page 11 in Hoff and Niu (2012).

    Parameters
    ----------
    m : real ndarray
        Column vector of shape (n x 1) of means in random effects model with 'n' being number of observations.
        Initialised in cov_reg_given_mean() with:

            m = (np.random.normal(0, 1, np.shape(y)[1])).reshape(-1, 1)

    v : real ndarray
        Column vector of shape (n x 1) of variances in random effects model with 'n' being number of observations.
        Initialised in cov_reg_given_mean() with:

            v = np.ones_like(m)

    x : real ndarray
        Matrix of shape (m x n) of covariates with 'm' being number of covariates
        and 'n' being number of observations.

    y : real ndarray
        Matrix of shape (p x n) of dependent variables with 'p' being number of dependent variables
        and 'n' being number of observations.

    basis : real ndarray
        Basis matrix used to estimate local mean - (A_est^T * basis) approximates local mean of y matrix.

    A_est : real ndarray
        Coefficient matrix used to estimate local mean - (A_est^T * basis) approximates local mean of y matrix.

    technique : string_like
        'direct' : Direct calculation method used in Hoff and Niu (2012).
            beta = [(x_tilda^T * x_tilda)^(-1)] * (x_tilda^T * y)

        'lasso' : Least Absolute Shrinkage and Selection Operator (LASSO) Regression.
            Minimize: (1 / (2 * n)) * ||y_tilda - x_tilda * beta||^2_2 +
                      alpha * ||beta||_1

        'ridge' :
            Minimize: ||y_tilda - x_tilda * beta||^2_2 + alpha * ||beta||^2_2
            Equivalent to: beta = [(x_tilda^T * x_tilda + alpha * I)^(-1)] * (x_tilda^T * y)

        'elastic-net' :
            Minimize: (1 / (2 * n)) * ||y_tilda - x_tilda * beta||^2_2 +
                      alpha * l1_ratio * ||beta||_1 + 0.5 * alpha * (1 - l1_ratio) * ||beta||^2_2

            l1_ratio = 1 equivalent to 'lasso'
            l1_ratio = 0 and alpha = 2 equivalent to 'ridge'

        'group-lasso' :
            With G being the grouping of the covariates the objective function is given below.
            Minimize: ||∑g∈G[X_g * beta_g] - y||^2_2 + alpha * ||w||_1 + lambda_group * ∑g∈G||beta_g||_2

        'sub-gradient' :
            Minimize: ||beta||_1
            subject to: x_tilda * beta^T = y
            iterate by: B_{k+1} = B_k - alpha_k(I_p - X^T * (X * X^T)^{-1} * X * sign(B_k))

    alpha : float
        Constant used in chosen regression to multiply onto weights.

    l1_ratio_or_reg : float
        Least Absolute Shrinkage and Selection Operator (LASSO) ratio for elastic-net regression and
        LASSO regulator for group LASSO regression.

    group_reg : float
        Group LASSO regulator for group LASSO regression.

    max_iter : positive integer
        Maximum number of iterations to perform in chosen regression.

    groups : real ndarray (integer ndarray)
        Groups to be used in 'group-lasso' regression.

    Returns
    -------
    B_est : real ndarray
        Coefficients for covariates explaining attributable covariance.

    Psi_est : real ndarray
        Base unattributable covariance present in model.

    Notes
    -----
    Group LASSO regression and Subgradient optimisation are experimental and need to be improved to stop possible
    breaking of correlation structure or nonsensical results.

    """

    if not isinstance(m, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('m must be of type np.ndarray.')
    if np.array(m).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('m must only contain floats.')
    try:
        if np.shape(np.array(m))[1] != 1:
            raise ValueError('m must be column vector.')
    except IndexError:
        raise ValueError('m must be column vector. Suggest: m.reshape(-1, 1)')
    if not isinstance(v, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('v must be of type np.ndarray.')
    if np.array(v).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('v must only contain floats.')
    try:
        if np.shape(np.array(v))[1] != 1:
            raise ValueError('v must be column vector.')
    except IndexError:
        raise ValueError('v must be column vector. Suggest: v.reshape(-1, 1)')

    if len(m) != len(v):
        raise ValueError('m and v are incompatible lengths.')

    if not isinstance(x, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('x must be of type np.ndarray.')
    if np.array(x).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('x must only contain floats.')

    if not isinstance(y, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('y must be of type np.ndarray.')
    if np.array(y).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('y must only contain floats.')

    if not isinstance(basis, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('basis must be of type np.ndarray.')
    if np.array(basis).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('basis must only contain floats.')

    if not isinstance(A_est, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('A_est must be of type np.ndarray.')
    if np.array(A_est).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('A_est must only contain floats.')

    if technique not in {'direct', 'lasso', 'ridge', 'elastic-net', 'group-lasso', 'sub-gradient'}:
        raise ValueError('\'technique\' not an acceptable value.')

    if (not isinstance(alpha, float)) or alpha < 0:
        raise ValueError('\'alpha\' must be a non-negative float.')
    if (not isinstance(l1_ratio_or_reg, float)) or l1_ratio_or_reg < 0:
        raise ValueError('\'l1_ratio_or_reg\' must be a non-negative float.')
    if (not isinstance(group_reg, float)) or group_reg < 0:
        raise ValueError('\'group_reg\' must be a non-negative float.')
    if (not isinstance(max_iter, int)) or max_iter < 1:
        raise ValueError('\'max_iter\' must be a positive integer.')
    try:
        if (groups is not None) and groups:
            if not isinstance(groups, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
                raise TypeError('groups must be of type np.ndarray.')
            if np.array(groups).dtype != np.array([[1, 1, 1, 1]]).dtype:
                raise TypeError('groups must only contain integers.')
    except ValueError:
        if any(groups):
            if not isinstance(groups, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
                raise TypeError('groups must be of type np.ndarray.')
            if np.array(groups).dtype != np.array([[1, 1, 1, 1]]).dtype:
                raise TypeError('groups must only contain integers.')

    x_tilda = np.vstack([m * x.T, (v ** (1 / 2)) * x.T])
    y_tilda = np.vstack(((y.T - np.matmul(A_est.T, basis).T), np.zeros_like(y.T)))

    if technique == 'direct':

        try:
            B_est = \
                np.matmul(y_tilda.T,
                          np.matmul(x_tilda, np.linalg.inv(np.matmul(x_tilda.T, x_tilda).astype(np.float64))))
        except:
            B_est = \
                np.matmul(y_tilda.T,
                          np.matmul(x_tilda, np.linalg.pinv(np.matmul(x_tilda.T, x_tilda).astype(np.float64))))

    elif technique == 'lasso':

        reg_lasso = linear_model.MultiTaskLasso(alpha=alpha, fit_intercept=False, max_iter=max_iter)
        reg_lasso.fit(x_tilda, y_tilda)
        B_est = reg_lasso.coef_

    elif technique == 'ridge':

        try:
            B_est = \
                np.matmul(np.matmul(y_tilda.T, x_tilda),
                          np.linalg.inv(np.matmul(x_tilda.T, x_tilda).astype(np.float64) +
                                        alpha * np.identity(np.shape(x_tilda)[1])))
        except:
            reg_ridge = linear_model.Ridge(alpha=alpha, fit_intercept=False, max_iter=max_iter)
            reg_ridge.fit(x_tilda, y_tilda)
            B_est = reg_ridge.coef_

    elif technique == 'elastic-net':

        reg_elas_net = linear_model.ElasticNet(alpha=alpha, fit_intercept=False, l1_ratio=l1_ratio_or_reg,
                                               max_iter=max_iter)
        reg_elas_net.fit(x_tilda, y_tilda)
        B_est = reg_elas_net.coef_

    elif technique == 'group-lasso':

        ################################################################
        # possibly breaks correlation structure when doing column-wise #
        ################################################################

        # https://group-lasso.readthedocs.io/en/latest/
        # https://group-lasso.readthedocs.io/en/latest/auto_examples/index.html
        B_est = np.zeros((np.shape(y_tilda)[1], np.shape(x_tilda)[1]))
        for covariate in range(np.shape(y_tilda)[1]):
            reg_group_lasso = group_lasso.GroupLasso(groups=groups, old_regularisation=True, supress_warning=True,
                                                     fit_intercept=False, group_reg=group_reg, l1_reg=l1_ratio_or_reg)
            reg_group_lasso.fit(x_tilda, y_tilda[:, covariate].reshape(-1, 1))
            B_est[covariate, :] = reg_group_lasso.coef_[:, 0]
            print(B_est[covariate, :])

    elif technique == 'sub-gradient':

        ################################################################
        # possibly breaks correlation structure when doing column-wise #
        ################################################################

        # B_est = np.zeros((np.shape(y_tilda)[1], np.shape(x_tilda)[1]))
        # for covariate in range(np.shape(y_tilda)[1]):
        #     # https://github.com/scikit-learn/scikit-learn/blob/main/sklearn/linear_model/_sgd_fast.pyx
        #     reg_sgd = SGDRegressor()
        #     reg_sgd.fit(x_tilda, y_tilda[:, covariate])
        #     B_est[covariate, :] = reg_sgd.coef_
        #     print(B_est[covariate, :])

        B_est = subgrad_opt(x_tilda, y_tilda, alpha=alpha, max_iter=max_iter)

    C_est = np.vstack((A_est, B_est.T))

    x_tilda_extend = np.hstack((np.vstack((basis.T, np.zeros_like(basis.T))), x_tilda))
    y_tilda_extend = np.vstack((y.T, np.zeros_like(y.T)))

    const = (y_tilda_extend - np.matmul(x_tilda_extend, C_est))
    try:
        Psi_est = np.matmul(const.T, const) / np.shape(x)[1]
    except:
        Psi_est = np.matmul(const.T, const) / len(x)

    return B_est.astype(np.float64), Psi_est.astype(np.float64)


def gamma_v_m_error(errors: np.ndarray, x: np.ndarray, Psi: np.ndarray, B: np.ndarray):

    if not isinstance(errors, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('errors must be of type np.ndarray.')
    if np.array(errors).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('errors must only contain floats.')
    if not isinstance(x, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('x must be of type np.ndarray.')
    if np.array(x).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('x must only contain floats.')
    if not isinstance(Psi, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('Psi must be of type np.ndarray.')
    if np.array(Psi).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('Psi must only contain floats.')
    if not isinstance(B, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('B must be of type np.ndarray.')
    if np.array(B).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('B must only contain floats.')

    """
    Function to calculate variance and mean for random error formulation which follows calculation
    at the bottom of page 9 in Hoff and Niu (2012).

    Parameters
    ----------
    errors : real ndarray
        Errors (variance) about given mean of dependent variables matrix.

    x : real ndarray
        Independent variable matrix.

    Psi : real ndarray
        Base unattributable covariance present in model.

    B : real ndarray
        Coefficients for covariates explaining attributable covariance.

    Returns
    -------
    m : real ndarray
        Mean of random error formulation.

    v : real ndarray
        Variance of random error formulation.

    Notes
    -----

    """
    try:
        const = np.matmul(np.linalg.solve(Psi.astype(np.float64), B.T.astype(np.float64)), x)
    except:
        try:
            const = np.matmul(np.linalg.lstsq(Psi.astype(np.float64), B.T.astype(np.float64), rcond=None)[0], x)
        except:
            const = np.matmul(np.linalg.lstsq(Psi.astype(np.float64).dot(Psi.astype(np.float64).T),
                                              Psi.astype(np.float64).dot(B.T.astype(np.float64)), rcond=None)[0], x)

    v = np.abs((1 + (x * np.matmul(B, const)).sum(0)) ** (-1))
    m = v * sum(errors * const)

    return m.astype(np.float64), v.astype(np.float64)


def cov_reg_given_mean(A_est: np.ndarray, basis: np.ndarray, x: np.ndarray, y: np.ndarray, iterations: int = 10,
                       technique: str = 'direct', alpha: float = 1.0, l1_ratio_or_reg: float = 0.1,
                       group_reg: float = 1e-6, max_iter: int = 10000, groups: np.ndarray = None):
    """
    Calculate Psi and B matrices of covariance regression as in Hoff and Niu (2012) except that A_est and basis
    are now given as inputs allowing for customisable definition of "mean" or "trend".

    Parameters
    ----------
    A_est : real ndarray
        Matrix of coefficients corresponding to 'basis' to estimate mean of dependent variables.

    basis : real ndarray
        Matrix of basis functions corresponding to 'A_est' to estimate mean of dependent variables.

    x : real ndarray
        Matrix of independent variables.

    y : real ndarray
        Matrix of dependent variables.

    iterations : positive integer
        Number of iterations of the Covariance Regression algorithm.

    technique : string
        'direct' : Direct calculation method used in Hoff and Niu (2012).
            beta = [(x_tild^T * x_tilda)^(-1)] * (x_tilda^T * y)

        'lasso' : Least Absolute Shrinkage and Selection Operator (LASSO) Regression.
            Minimize: (1 / (2 * n)) * ||y_tilda - x_tilda * beta||^2_2 +
                      alpha * ||beta||_1

        'ridge' :
            Minimize: ||y_tilda - x_tilda * beta||^2_2 + alpha * ||beta||^2_2
            Equivalent to: beta = [(x_tild^T * x_tilda + alpha * I)^(-1)] * (x_tilda^T * y)

        'elastic-net' :
            Minimize: (1 / (2 * n)) * ||y_tilda - x_tilda * beta||^2_2 +
                      alpha * l1_ratio * ||beta||_1 + 0.5 * alpha * (1 - l1_ratio) * ||beta||^2_2

            l1_ratio = 1 equivalent to 'lasso'
            l1_ratio = 0 and alpha = 2 equivalent to 'ridge'

        'group-lasso' :
            With G being the grouping of the covariates the objective function is given below.
            Minimize: ||∑g∈G[X_g * beta_g] - y||^2_2 + alpha * ||w||_1 + lambda_group * ∑g∈G||beta_g||_2

        'sub-gradient' :
            Minimize: ||beta||_1
            subject to: x_tilda * beta^T = y
            iterate by: B_{k+1} = B_k - alpha_k(I_p - X^T * (X * X^T)^{-1} * X * sign(B_k))

    alpha : float
        Lambda value used to weight coefficients.

    l1_ratio_or_reg : float
        Ratio of l1 normalisation in elastic-net regression.

    group_reg : float
        Lambda weighting for group lasso regression.

    max_iter : positive integer
        Maximum number of iterations in regularised regression.

    groups : real ndarray (consisting of non-negative integers)
        Vector of groups to be used in group LASSO regression.

    Returns
    -------
    B_est : real ndarray
        Coefficients for covariates explaining attributable covariance.

    Psi_est : real ndarray
        Base unattributable covariance present in model.

    Notes
    -----

    """
    if not isinstance(A_est, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('A_est must be of type np.ndarray.')
    if np.array(A_est).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('A_est must only contain floats.')

    if not isinstance(basis, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('basis must be of type np.ndarray.')
    if np.array(basis).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('basis must only contain floats.')

    if not isinstance(x, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('x must be of type np.ndarray.')
    if np.array(x).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('x must only contain floats.')

    if not isinstance(y, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('y must be of type np.ndarray.')
    if np.array(y).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('y must only contain floats.')

    if (not isinstance(iterations, int)) or iterations < 1:
        raise ValueError('\'iterations\' must be a positive integer.')
    if technique not in {'direct', 'lasso', 'ridge', 'elastic-net', 'group-lasso', 'sub-gradient'}:
        raise ValueError('\'technique\' not an acceptable value.')
    if (not isinstance(alpha, float)) or alpha < 0:
        raise ValueError('\'alpha\' must be a non-negative float.')
    if (not isinstance(l1_ratio_or_reg, float)) or l1_ratio_or_reg < 0:
        raise ValueError('\'l1_ratio_or_reg\' must be a non-negative float.')
    try:
        if (groups is not None) and groups:
            if not isinstance(groups, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
                raise TypeError('groups must be of type np.ndarray.')
            if np.array(groups).dtype != np.array([[1, 1, 1, 1]]).dtype:
                raise TypeError('groups must only contain integers.')
    except ValueError:
        if any(groups):
            if not isinstance(groups, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
                raise TypeError('groups must be of type np.ndarray.')
            if np.array(groups).dtype != np.array([[1, 1, 1, 1]]).dtype:
                raise TypeError('groups must only contain integers.')

    if groups is None:
        groups = {}
    m = (np.random.normal(0, 1, np.shape(y)[1])).reshape(-1, 1)  # initialise m
    v = np.ones_like(m)  # initialise v

    mean = np.matmul(A_est.T, basis)

    for iter in range(iterations):

        B_est, Psi_est = calc_B_Psi(m=m, v=v, x=x, y=y, basis=basis, A_est=A_est, technique=technique,
                                    l1_ratio_or_reg=l1_ratio_or_reg, group_reg=group_reg, alpha=alpha,
                                    max_iter=max_iter, groups=groups)

        m, v = gamma_v_m_error(errors=(y - mean), x=x, Psi=Psi_est, B=B_est.T)
        m = m.reshape(-1, 1)
        v = v.reshape(-1, 1)

    B_est = B_est.T

    return B_est.astype(np.float64), Psi_est.astype(np.float64)


def subgrad_opt(x_tilda: np.ndarray, y_tilda: np.ndarray, max_iter: int, alpha: float = 1e-12):
    """
    Subgradient optimisation of coefficients.

    Parameters
    ----------
    x_tilda : real ndarray
        Matrix of independent variables.

    y_tilda : real ndarray
        Matrix of dependent variables.

    max_iter : positive integer
        Maximum number of integers.

    alpha : float
        Scale to be used in square summable, but not summable Polyak step size.

    Returns
    -------
    B_est : real ndarray
        Coefficients for covariates explaining attributable covariance.

    Notes
    -----
    Convergence results do not apply if applied to skinny matrices.
    Starting point of algorithm can be changed and optimised.

    """

    if not isinstance(x_tilda, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('x_tilda must be of type np.ndarray.')
    if np.array(x_tilda).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('x_tilda must only contain floats.')

    if not isinstance(y_tilda, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('y_tilda must be of type np.ndarray.')
    if np.array(y_tilda).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('y_tilda must only contain floats.')

    if (not isinstance(max_iter, int)) or max_iter < 1:
        raise ValueError('\'max_iter\' must be a positive integer.')
    if (not isinstance(alpha, float)) or alpha < 0:
        raise ValueError('\'alpha\' must be a non-negative float.')

    # will not necessarily converge if not satisfied
    # if np.shape(x_tilda)[0] > np.shape(x_tilda)[1]:
    #     raise ValueError('Matrix cannot be skinny/thin.')

    # reg_ridge = linear_model.Ridge(alpha=alpha, fit_intercept=False)
    # reg_ridge.fit(x_tilda, y_tilda)
    # B_sol = reg_ridge.coef_.T

    B_sol = np.matmul(y_tilda.T, np.matmul(x_tilda, np.linalg.pinv(np.matmul(x_tilda.T, x_tilda).astype(np.float64)))).T
    B_k = B_sol.copy()

    f_star = sum(np.abs(B_sol))
    f_best = 1e12 * np.ones_like(f_star)

    k = 1

    while k < int(max_iter + 1):

        B_k_1 = B_k - (alpha / k) * np.matmul((np.identity(np.shape(x_tilda)[1]) -
                                             np.matmul(np.matmul(x_tilda.T,
                                                                 np.linalg.pinv(np.matmul(x_tilda, x_tilda.T))),
                                                       x_tilda)), np.sign(B_k))
        f_potential = sum(np.abs(B_k_1))

        for i in range(len(f_potential)):
            if f_potential[i] <= f_best[i]:
                f_best[i] = f_potential[i]
                B_k[:, i] = B_k_1[:, i]

        print(k)
        k += 1

    B_est = B_k.copy().T

    return B_est




#     ________
#            /
#      \    /
#       \  /
#        \/

# Document Strings Publication

# Reference work:

# Bauwens, L., Laurent, S. and Rombouts, J.,
# Multivariate GARCH Models: A Survey. Journal of Applied
# Econometrics, 2006, 21, 79–109.

# Log-likelihood reference work - equations referenced throughout:

# Engle, R., Dynamic Conditional Correlation: A Simple Class of
# Multivariate Generalized Autoregressive Conditional Heteroskedasticity Models.
# Journal of Business & Economic Statistics, 2002, 20, 339–350.

# This package created as the following Python package did not work for our purposes:
# mgarch 0.2.0 - available at https://pypi.org/project/mgarch/

import numpy as np
import pandas as pd
from arch import arch_model
from scipy.optimize import minimize


def covregpy_dcc(returns_matrix, p=3, q=3, days=10, print_correlation=False, rescale=False):
    """
    Dynamic Conditional Correlation - Multivariate Generalized Autoregressive Conditional Heteroskedasticity model.

    Parameters
    ----------
    returns_matrix : real ndarray
        Matrix of asset returns.

    p : positive integer
        GARCH(p, q) model 'p' value used in calculation of uni-variate variance used in model.

    q : positive integer
        GARCH(p, q) model 'q' value used in calculation of uni-variate variance used in model.

    days : positive integer
        Number of days ahead to forecast covariance.

    print_correlation : boolean
        Debugging to gauge if correlation structure is reasonable.

    rescale : boolean
        Automatically rescale data to help if encountering convergence problems - see:
        https://arch.readthedocs.io/en/latest/univariate/introduction.html

    Returns
    -------
    forecasted_covariance : real ndarray
        Forecasted covariance.

    Notes
    -----

    """
    if not isinstance(returns_matrix, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])),
                                       type(pd.DataFrame(np.asarray([[1.0, 2.0], [3.0, 4.0]]))))):
        raise TypeError('Returns must be of type np.ndarray and pd.Dataframe.')
    if pd.isnull(np.asarray(returns_matrix)).any():
        raise TypeError('Returns must not contain nans.')
    if np.array(returns_matrix).dtype != np.array([[1., 1.], [1., 1.]]).dtype:
        raise TypeError('Returns must only contain floats.')
    if (not isinstance(p, int)) or (p <= 0):
        raise ValueError('\'p\' must be a positive integer.')
    if (not isinstance(q, int)) or (q <= 0):
        raise ValueError('\'q\' must be a positive integer.')
    if (not isinstance(days, int)) or (days <= 0):
        raise ValueError('\'days\' must be a positive integer.')
    if not isinstance(print_correlation, bool):
        raise TypeError('\'print_correlation\' must be boolean.')
    if not isinstance(rescale, bool):
        raise TypeError('\'rescale\' must be boolean.')

    # convert Dataframe
    returns_matrix = np.asarray(returns_matrix)

    # flip matrix to be consistent
    if np.shape(returns_matrix)[0] < np.shape(returns_matrix)[1]:
        returns_matrix = returns_matrix.T

    # initialised modelled variance matrix
    modelled_variance = np.zeros_like(returns_matrix)

    # iteratively calculate modelled variance using univariate GARCH model
    for stock in range(np.shape(returns_matrix)[1]):
        model = arch_model(returns_matrix[:, stock], mean='Zero', vol='GARCH', p=p, q=q, rescale=rescale)
        model_fit = model.fit(disp="off")
        # if rescale:
        #     modelled_variance[:, stock] = model_fit.conditional_volatility / model.scale
        # else:
        #     modelled_variance[:, stock] = model_fit.conditional_volatility
        modelled_variance[:, stock] = model_fit.conditional_volatility

    # optimise alpha & beta parameters to be used in page 90 equation (40)
    params = minimize(dcc_loglike, (0.01, 0.94),
                      args=(returns_matrix, modelled_variance),
                      bounds=((1e-6, 1), (1e-6, 1)),
                      method='Nelder-Mead')

    # list of optimisation methods available

    # Nelder-Mead, Powell, CG, BFGS, Newton-CG,
    # L-BFGS-B, TNC, COBYLA, SLSQP, trust-constr,
    # dogleg, trust-ncg, trust-exact, trust-krylov

    a = params.x[0]
    b = params.x[1]

    t = np.shape(returns_matrix)[0]  # time interval
    q_bar = np.cov(returns_matrix.T)  # base (unconditional) covariance

    # setup matrices
    q_t = np.zeros((t, np.shape(q_bar)[0], np.shape(q_bar)[1]))  # page 90 - Equation (40)
    h_t = np.zeros((t, np.shape(q_bar)[0], np.shape(q_bar)[1]))  # page 89 - Equation (35)
    dts = np.zeros((t, np.shape(q_bar)[0], np.shape(q_bar)[1]))  # page 88 - Equation (32)
    dts_inv = np.zeros((t, np.shape(q_bar)[0], np.shape(q_bar)[1]))  # used in calculation of u_t
    u_t = np.zeros((np.shape(q_bar)[0], t))  # page 89 - defined between Equation (37) and (38)
    qts = np.zeros((t, np.shape(q_bar)[0], np.shape(q_bar)[1]))  # page 90 - defined within Equation (39)
    r_t = np.zeros((t, np.shape(q_bar)[0], np.shape(q_bar)[1]))  # page 90 - Equation (39)

    # initialise q_t
    # unsure of initialisation
    q_t[0] = np.zeros_like(np.matmul(returns_matrix[0, :].reshape(-1, 1) / 2, returns_matrix[0, :].reshape(1, -1) / 2))

    for var in range(t):
        # page 88 - Equation (32)
        dts[var] = np.diag(modelled_variance[int(var - 1), :])  # modelled variance - page 89 - Equation (34)
        # page 89 - defined between Equation (37) and (38)
        try:
            dts_inv[var] = np.linalg.inv(dts[var])
        except:
            dts_inv[var] = np.linalg.pinv(dts[var])
        u_t[:, var] = np.matmul(dts_inv[var], returns_matrix[var, :].reshape(-1, 1))[:, 0]

    for var in range(1, t):

        # page 90 - Equation (40)
        q_t[var] = (1 - a - b) * q_bar + \
                   a * np.matmul(u_t[:, int(var - 1)].reshape(-1, 1), u_t[:, int(var - 1)].reshape(1, -1)) + \
                   b * q_t[int(var - 1)]

        # page 90 - defined within Equation (39)
        try:
            qts[var] = np.linalg.inv(np.sqrt(np.diag(np.diag(q_t[var]))))
        except:
            qts[var] = np.linalg.pinv(np.sqrt(np.diag(np.diag(q_t[var]))))

        # page 90 - Equation (39)
        r_t[var] = np.matmul(qts[var], np.matmul(q_t[var], qts[var]))

        # page 89 - Equation (35)
        h_t[var] = np.matmul(dts[var], np.matmul(r_t[var], dts[var]))

    # Brownian uncertainty with  variance increasing linearly with square root of time
    forecasted_covariance = h_t[-1] * np.sqrt(days)

    if print_correlation:
        corr = np.zeros_like(forecasted_covariance)
        for i in range(np.shape(corr)[0]):
            for j in range(np.shape(corr)[1]):
                corr[i, j] = forecasted_covariance[i, j] / (np.sqrt(forecasted_covariance[i, i]) *
                                                            np.sqrt(forecasted_covariance[j, j]))
        print(corr)

    return forecasted_covariance


def dcc_loglike(params, returns_matrix, modelled_variance):
    """
    Dynamic Conditional Correlation loglikelihood function to optimise.

    Parameters
    ----------
    params : real ndarray
        Real array of shape (2, ) with two parameters to optimise.

    returns_matrix : real ndarray
        Matrix of asset returns.

    modelled_variance : real ndarray
        Uni-variate modelled variance to be used in optimisation.

    Returns
    -------
    loglike : float
        Loglikelihood function to be minimised.

    Notes
    -----

    """
    if pd.isnull(np.asarray(params)).any():
        raise ValueError('Parameters must not contain nans.')
    if np.array(params).dtype != np.array([[1., 1.], [1., 1.]]).dtype:
        raise ValueError('Parameters must only contain floats.')
    if pd.isnull(np.asarray(returns_matrix)).any():
        raise ValueError('Returns must not contain nans.')
    if np.array(returns_matrix).dtype != np.array([[1., 1.], [1., 1.]]).dtype:
        raise ValueError('Returns must only contain floats.')
    if pd.isnull(np.asarray(modelled_variance)).any():
        raise ValueError('Covariance must not contain nans.')
    if np.array(modelled_variance).dtype != np.array([[1., 1.], [1., 1.]]).dtype:
        raise ValueError('Covariance must only contain floats.')

    t = np.shape(returns_matrix)[0]  # time interval
    q_bar = np.cov(returns_matrix.T)  # base (unconditional) covariance

    # setup matrices
    q_t = np.zeros((t, np.shape(q_bar)[0], np.shape(q_bar)[1]))  # page 90 - Equation (40)
    h_t = np.zeros((t, np.shape(q_bar)[0], np.shape(q_bar)[1]))  # page 89 - Equation (35)
    dts = np.zeros((t, np.shape(q_bar)[0], np.shape(q_bar)[1]))  # page 88 - Equation (32)
    dts_inv = np.zeros((t, np.shape(q_bar)[0], np.shape(q_bar)[1]))  # used in calculation of u_t
    u_t = np.zeros((np.shape(q_bar)[0], t))  # page 89 - defined between Equation (37) and (38)
    qts = np.zeros((t, np.shape(q_bar)[0], np.shape(q_bar)[1]))  # page 90 - defined within Equation (39)
    r_t = np.zeros((t, np.shape(q_bar)[0], np.shape(q_bar)[1]))  # page 90 - Equation (39)

    # initialise q_t
    # unsure of initialisation
    q_t[0] = np.zeros_like(np.matmul(returns_matrix[0, :].reshape(-1, 1) / 2, returns_matrix[0, :].reshape(1, -1) / 2))

    for var in range(t):
        # page 88 - Equation (32)
        dts[var] = np.diag(modelled_variance[int(var - 1), :])  # modelled variance - page 89 - Equation (34)
        # page 89 - defined between Equation (37) and (38)
        try:
            dts_inv[var] = np.linalg.inv(dts[var])
        except:
            dts_inv[var] = np.linalg.pinv(dts[var])
        u_t[:, var] = np.matmul(dts_inv[var], returns_matrix[var, :].reshape(-1, 1))[:, 0]

    # initialise log-likehood value
    loglike = 0

    for var in range(1, t):

        # page 90 - Equation (40)
        q_t[var] = (1 - params[0] - params[1]) * q_bar + \
                   params[0] * np.matmul(u_t[:, int(var - 1)].reshape(-1, 1), u_t[:, int(var - 1)].reshape(1, -1)) + \
                   params[1] * q_t[int(var - 1)]

        # page 90 - defined within Equation (39)
        try:
            qts[var] = np.linalg.inv(np.sqrt(np.diag(np.diag(q_t[var]))))
        except:
            qts[var] = np.linalg.pinv(np.sqrt(np.diag(np.diag(q_t[var]))))

        # page 90 - Equation (39)
        r_t[var] = np.matmul(qts[var], np.matmul(q_t[var], qts[var]))

        # page 89 - Equation (35)
        h_t[var] = np.matmul(dts[var], np.matmul(r_t[var], dts[var]))

        # likelihood function from reference work page 11 - Equation 26
        try:
            loglike -= (np.shape(q_bar)[0] * np.log(2 * np.pi) +
                        2 * np.log(np.linalg.det(dts[var])) + np.log(np.linalg.det(r_t[var])) +
                        np.matmul(u_t[:, var].reshape(1, -1),
                                  np.matmul(np.linalg.inv(r_t[var]),
                                            u_t[:, var].reshape(-1, 1)))[0][0])
        except:
            loglike -= (np.shape(q_bar)[0] * np.log(2 * np.pi) +
                        2 * np.log(np.linalg.det(dts[var])) + np.log(np.linalg.det(r_t[var])) +
                        np.matmul(u_t[:, var].reshape(1, -1),
                                  np.matmul(np.linalg.pinv(r_t[var]),
                                            u_t[:, var].reshape(-1, 1)))[0][0])

    return loglike




#     ________
#            /
#      \    /
#       \  /
#        \/

# Main reference: Bonizzi, Karel, Meste, & Peeters (2014)
# Bonizzi, P., Karel, J., Meste, O., & Peeters, R. (2014).
# Singular Spectrum Decomposition: A New Method for Time Series Decomposition.
# Advances in Adaptive Data Analysis, 6(04), 1450011 (1-34). World Scientific.

import scipy as sp
import numpy as np
import seaborn as sns
from matplotlib import mlab
import matplotlib.pyplot as plt
from scipy.optimize import minimize
from CovRegpy_SSA import CovRegpy_ssa
from scipy.fft import fft

np.random.seed(0)

sns.set(style='darkgrid')


def gaussian(f, A, mu, sigma):
    """
    Gaussian distribution to be fitted to power-spectral density.

    Parameters
    ----------
    f : real ndarray
        Frequency over which distribution will be fitted.

    A : float
        Amplitude of Gaussian distribution.

    mu : float
        Mean of Gaussian distribution.

    sigma : float
        Standard deviation of Gaussian distribution.

    Returns
    -------
    distribution : real ndarray
        Gaussian distribution over domain 'f'.

    Notes
    -----
    Used in calculation of downsampling range.

    """

    if not isinstance(f, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'f\' must be of type np.ndarray.')
    if np.array(f).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'f\' must only contain floats.')
    try:
        if (not isinstance(A[0], float)) or A[0] < 0:
            raise ValueError('\'A\' must be a non-negative float.')
        if (not isinstance(mu[0], float)) or mu[0] <= 0:
            raise ValueError('\'mu\' must be a positive float.')
        if (not isinstance(sigma[0], float)) or sigma[0] <= 0:
            raise ValueError('\'sigma\' must be a positive float.')
    except (IndexError, TypeError):
        if (not isinstance(A, float)) or A < 0:
            raise ValueError('\'A\' must be a non-negative float.')
        if (not isinstance(mu, float)) or mu <= 0:
            raise ValueError('\'mu\' must be a positive float.')
        if (not isinstance(sigma, float)) or sigma <= 0:
            raise ValueError('\'sigma\' must be a positive float.')

    distribution = A * np.exp(- (f - mu) ** 2 / (2 * sigma ** 2))

    return distribution


def max_bool(time_series):
    """
    Calculate maximum boolean of time series.

    Parameters
    ----------
    time_series : real ndarray
        Time series where maximums are to be calculated.

    Returns
    -------
    max_bool : real ndarray
        Boolean array with location of maximums.

    Notes
    -----
    Used to calculate centres of each Gaussian distributions.

    """

    if not isinstance(time_series, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'time_series\' must be of type np.ndarray.')
    if np.array(time_series).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'time_series\' must only contain floats.')

    max_bool = np.r_[False, time_series[1:] >= time_series[:-1]] & np.r_[time_series[:-1] > time_series[1:], False]

    return max_bool


def spectral_obj_func_l1(theta, f, mu_1, mu_2, mu_3, spectrum):
    """
    Calculate 3 Gaussian functions.

    Parameters
    ----------
    theta : real ndarray
        A vector of shape (6,) containing amplitudes and sigmas.

    f : real ndarray
        Frequency over which Gaussian distributions and frequency spectrum are fitted.

    mu_1 : float
        Mean of the highest frequency peak.

    mu_2 : float
         Mean of the second-highest frequency peak.

    mu_3 : float
        Mean of remaining spectra.

    spectrum : real ndarray
        Spectrum to be fitted to Gaussian distributions.

    Returns
    -------
    objective_function : real ndarray
        The l1 norm of difference between spectrum and Gaussian distributions.

    Notes
    -----
    Should experiment with different norms.

    """

    if not isinstance(theta, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'theta\' must be of type np.ndarray.')
    if np.array(theta).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'theta\' must only contain floats.')
    if len(theta) != 6:
        raise ValueError('\'theta\' must be of length 6.')
    if not isinstance(f, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'f\' must be of type np.ndarray.')
    if np.array(f).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'f\' must only contain floats.')
    try:
        if (not isinstance(mu_1[0], float)) or mu_1[0] <= 0:
            raise ValueError('\'mu_1\' must be a positive float.')
        if (not isinstance(mu_2[0], float)) or mu_2[0] <= 0:
            raise ValueError('\'mu_2\' must be a positive float.')
        if (not isinstance(mu_3[0], float)) or mu_3[0] <= 0:
            raise ValueError('\'mu_3\' must be a positive float.')
    except (IndexError, TypeError):
        if (not isinstance(mu_1, float)) or mu_1 <= 0:
            raise ValueError('\'mu_1\' must be a positive float.')
        if (not isinstance(mu_2, float)) or mu_2 <= 0:
            raise ValueError('\'mu_2\' must be a positive float.')
        if (not isinstance(mu_3, float)) or mu_3 <= 0:
            raise ValueError('\'mu_3\' must be a positive float.')
    if not isinstance(spectrum, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'spectrum\' must be of type np.ndarray.')
    if np.array(spectrum).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'spectrum\' must only contain floats.')
    if len(f) != len(spectrum):
        raise ValueError('\'f\' and \'spectrum\' are incompatible lengths.')

    objective_function = sum(np.abs(theta[0] * np.exp(- (f - mu_1) ** 2 / (2 * theta[3] ** 2)) +
                                    theta[1] * np.exp(- (f - mu_2) ** 2 / (2 * theta[4] ** 2)) +
                                    theta[2] * np.exp(- (f - mu_3) ** 2 / (2 * theta[5] ** 2)) - spectrum))

    return objective_function


def spectral_obj_func_l2(theta, f, mu_1, mu_2, mu_3, spectrum):
    """
    Calculate 3 Gaussian functions.

    Parameters
    ----------
    theta : real ndarray
        A vector of shape (6,) containing amplitudes and sigmas.

    f : real ndarray
        Frequency over which Gaussian distributions and frequency spectrum are fitted.

    mu_1 : float
        Mean of the highest frequency peak.

    mu_2 : float
         Mean of the second-highest frequency peak.

    mu_3 : float
        Mean of remaining spectra.

    spectrum : real ndarray
        Spectrum to be fitted to Gaussian distributions.

    Returns
    -------
    objective_function : real ndarray
        The l2 norm of difference between spectrum and Gaussian distributions.

    Notes
    -----
    Should experiment with different norms.

    """

    if not isinstance(theta, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'theta\' must be of type np.ndarray.')
    if np.array(theta).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'theta\' must only contain floats.')
    if len(theta) != 6:
        raise ValueError('\'theta\' must be of length 6.')
    if not isinstance(f, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'f\' must be of type np.ndarray.')
    if np.array(f).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'f\' must only contain floats.')
    try:
        if (not isinstance(mu_1[0], float)) or mu_1[0] <= 0:
            raise ValueError('\'mu_1\' must be a positive float.')
        if (not isinstance(mu_2[0], float)) or mu_2[0] <= 0:
            raise ValueError('\'mu_2\' must be a positive float.')
        if (not isinstance(mu_3[0], float)) or mu_3[0] <= 0:
            raise ValueError('\'mu_3\' must be a positive float.')
    except (IndexError, TypeError):
        if (not isinstance(mu_1, float)) or mu_1 <= 0:
            raise ValueError('\'mu_1\' must be a positive float.')
        if (not isinstance(mu_2, float)) or mu_2 <= 0:
            raise ValueError('\'mu_2\' must be a positive float.')
        if (not isinstance(mu_3, float)) or mu_3 <= 0:
            raise ValueError('\'mu_3\' must be a positive float.')
    if not isinstance(spectrum, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'spectrum\' must be of type np.ndarray.')
    if np.array(spectrum).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'spectrum\' must only contain floats.')
    if len(f) != len(spectrum):
        raise ValueError('\'f\' and \'spectrum\' are incompatible lengths.')

    objective_function = sum((theta[0] * np.exp(- (f - mu_1) ** 2 / (2 * theta[3] ** 2)) +
                              theta[1] * np.exp(- (f - mu_2) ** 2 / (2 * theta[4] ** 2)) +
                              theta[2] * np.exp(- (f - mu_3) ** 2 / (2 * theta[5] ** 2)) - spectrum) ** 2)

    return objective_function


def constraint_positive(x):
    """
    Constraint on amplitudes and sigmas being positive.

    Parameters
    ----------
    x : real ndarray
        Parameters to be constrained to be positive.

    Returns
    -------
    x : real ndarray
        Positive constrained parameters.

    Notes
    -----
    Constraint function.

    """
    return x


def gaus_param(w0, f, mu_1, mu_2, mu_3, spectrum, method='l1'):
    """
    Function that calculates optimal amplitudes and sigmas.

    Parameters
    ----------
    w0 : real ndarray
        Vectors of starting parameters.

    f : real ndarray
        Frequency over which parameters are optimised.

    mu_1 : float
        Mean of the highest frequency peak.

    mu_2 : float
         Mean of the second-highest frequency peak.

    mu_3 : float
        Mean of remaining spectra.

    spectrum : real ndarray
        Spectrum to be fitted to Gaussian distributions.

    method : string
        Whether to us l1 norm or l2 norm.

    Returns
    -------
    OptimizeResult : OptimizeResult
        Returns optimised objective function as well as parameters.

    Notes
    -----

    """

    if not isinstance(w0, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'w0\' must be of type np.ndarray.')
    if np.array(w0).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'w0\' must only contain floats.')
    if len(w0) != 6:
        raise ValueError('\'w0\' must be of length 6.')
    if not isinstance(f, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'f\' must be of type np.ndarray.')
    if np.array(f).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'f\' must only contain floats.')
    try:
        if (not isinstance(mu_1[0], float)) or mu_1[0] <= 0:
            raise ValueError('\'mu_1\' must be a positive float.')
        if (not isinstance(mu_2[0], float)) or mu_2[0] <= 0:
            raise ValueError('\'mu_2\' must be a positive float.')
        if (not isinstance(mu_3[0], float)) or mu_3[0] <= 0:
            raise ValueError('\'mu_3\' must be a positive float.')
    except (IndexError, TypeError):
        if (not isinstance(mu_1, float)) or mu_1 <= 0:
            raise ValueError('\'mu_1\' must be a positive float.')
        if (not isinstance(mu_2, float)) or mu_2 <= 0:
            raise ValueError('\'mu_2\' must be a positive float.')
        if (not isinstance(mu_3, float)) or mu_3 <= 0:
            raise ValueError('\'mu_3\' must be a positive float.')
    if not isinstance(spectrum, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'spectrum\' must be of type np.ndarray.')
    if np.array(spectrum).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'spectrum\' must only contain floats.')
    if len(f) != len(spectrum):
        raise ValueError('\'f\' and \'spectrum\' are incompatible lengths.')
    if method not in {'l1', 'l2'}:
        raise ValueError('\'method\' not an acceptable value.')

    cons = ({'type': 'ineq', 'fun': constraint_positive})
    if method == 'l1':
        return minimize(spectral_obj_func_l1, x0=w0,
                        args=(f, mu_1, mu_2, mu_3, spectrum), method='nelder-mead',
                        constraints=cons, bounds=[(0, None), (0, None), (0, None), (0, None), (0, None), (0, None)])
    elif method == 'l2':
        return minimize(spectral_obj_func_l2, x0=w0,
                        args=(f, mu_1, mu_2, mu_3, spectrum), method='nelder-mead',
                        constraints=cons, bounds=[(0, None), (0, None), (0, None), (0, None), (0, None), (0, None)])


def scaling_factor_obj_func(a, residual_time_series, trend_estimate):
    """
    Scaling factor calculation.

    Parameters
    ----------
    a : float
        Scaling factor to be optimised.

    residual_time_series : real ndarray
        Residual time series to which scaled trend estimate is fitted.

    trend_estimate : real ndarray
        Component estimate to be fitted with scaling factor.

    Returns
    -------
    l2_error : float
        L2 error of scaled result and residual time series.

    Notes
    -----

    """

    try:
        if (not isinstance(a[0], float)) or a[0] <= 0:
            raise ValueError('\'a\' must be a positive float.')
    except (IndexError, TypeError):
        if (not isinstance(a, float)) or a <= 0:
            raise ValueError('\'a\' must be a positive float.')
    if not isinstance(residual_time_series, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'residual_time_series\' must be of type np.ndarray.')
    if np.array(residual_time_series).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'residual_time_series\' must only contain floats.')
    if not isinstance(trend_estimate, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'trend_estimate\' must be of type np.ndarray.')
    if np.array(trend_estimate).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'trend_estimate\' must only contain floats.')
    if len(residual_time_series) != len(trend_estimate):
        raise ValueError('\'residual_time_series\' and \'trend_estimate\' are incompatible lengths.')


    l2_error = sum((residual_time_series - a * trend_estimate) ** 2)

    return l2_error


def scaling_factor(residual_time_series, trend_estimate):
    """
    Scaling factor calculation.

    Parameters
    ----------
    residual_time_series : real ndarray
        Residual time series to which scaled trend estimate is fitted.

    trend_estimate : real ndarray
        Component estimate to be fitted with scaling factor.

    Returns
    -------
    OptimizeResult : OptimizeResult
        Optimised result and optimised scale factor.

    Notes
    -----

    """

    if not isinstance(residual_time_series, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'residual_time_series\' must be of type np.ndarray.')
    if np.array(residual_time_series).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'residual_time_series\' must only contain floats.')
    if not isinstance(trend_estimate, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'trend_estimate\' must be of type np.ndarray.')
    if np.array(trend_estimate).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'trend_estimate\' must only contain floats.')
    if len(residual_time_series) != len(trend_estimate):
        raise ValueError('\'residual_time_series\' and \'trend_estimate\' are incompatible lengths.')

    cons = ({'type': 'ineq', 'fun': constraint_positive})
    return minimize(scaling_factor_obj_func, x0=np.asarray(1),
                    args=(residual_time_series, trend_estimate),
                    method='nelder-mead', constraints=cons)


def CovRegpy_ssd(time_series, initial_trend_ratio=3.0, nmse_threshold=0.01, plot=False, debug=False, method='l2'):
    """
    Singular Spectrum Decomposition based on Bonizzi, Karel, Meste, & Peeters (2014).

    Parameters
    ----------
    time_series : real ndarray
        Time series to decompose.

    initial_trend_ratio : positive float
        Recommended initial_trend_ratio=3.0 in Bonizzi, Karel, Meste, & Peeters (2014).
        Not appropriate for some trends as demonstrated in example.

    nmse_threshold : positive float
        Normalised Mean-Squared Error stopping criterion from Bonizzi, Karel, Meste, & Peeters (2014).
        Should explore additional stopping criteria.

    plot : boolean
        Plot all intermediate stages as part of debugging.

    debug : boolean
        Print incremental residual nmse value - while loop continues until drops below threshold of nmse_threshold.

    method : string
        Whether to use l1 or l2 optimisation in gaus_param() function.

    Returns
    -------
    time_series_est_mat : real ndarray
        Matrix containing decomposition of time series - iterates until nmse_threshold is met.

    Notes
    -----
    Many expansions possible. We explore some of these in attempt to make more robust.

    """

    if not isinstance(time_series, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('\'time_series\' must be of type np.ndarray.')
    if np.array(time_series).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('\'time_series\' must only contain floats.')
    if (not isinstance(initial_trend_ratio, float)) or initial_trend_ratio <= 0:
        raise ValueError('\'initial_trend_ratio\' must be a positive float.')
    if (not isinstance(nmse_threshold, float)) or nmse_threshold <= 0:
        raise ValueError('\'nmse_threshold\' must be a positive float.')
    if not isinstance(plot, bool):
        raise TypeError('\'plot\' must be boolean.')
    if not isinstance(debug, bool):
        raise TypeError('\'debug\' must be boolean.')
    if method not in {'l1', 'l2'}:
        raise ValueError('\'method\' not an acceptable value.')

    # make duplicate of original time series
    time_series_orig = time_series.copy()
    time_series_resid = time_series.copy()

    # first test for 'SIZEABLE' trend
    dt = (24 * 60 * 60)
    s, f = mlab.psd(np.asarray(time_series_resid), Fs=1 / dt)
    if plot:
        plt.title('Power Spectral Density of Original Time Series')
        plt.plot(f * dt, s)
        plt.show()

    # if power is concentrated at first or second frequency band i.e. if time series is transient in nature.
    if np.log10(s)[1] == np.max(np.log10(s)) or np.log10(s)[0] == np.max(np.log10(s)):
        trend_est = \
            CovRegpy_ssa(time_series=np.asarray(time_series_resid),
                         L=int(len(np.asarray(time_series_resid)) / initial_trend_ratio), est=1)[0]
        if plot:
            plt.title('Original Transient Time Series and Trend Estimate')
            plt.plot(np.asarray(time_series_resid), label='Transient time series')
            plt.plot(trend_est, '--', label='Trend estimate')
            plt.legend(loc='best')
            plt.show()
        time_series_resid -= trend_est
        if plot:
            plt.title('Residual Time Series after Initial Transient Trend Removed')
            plt.plot(np.asarray(time_series_resid))
            plt.show()
            s, f = mlab.psd(np.asarray(time_series_resid), Fs=1 / dt)
            plt.title('Power Spectral Density after Initial Transient Trend Removed')
            plt.plot(f * dt, s)
            plt.show()

    try:
        time_series_est_mat = trend_est.reshape(1, -1)
    except:
        pass

    # Ensure initial nmse value initialises while loop
    nmse = nmse_threshold + 0.01

    while nmse > nmse_threshold:
        if debug:
            print(f'Normalised Mean-Squared Error: {nmse}')

        s, f = mlab.psd(np.asarray(time_series_resid), Fs=1 / dt)

        mu_1 = f[s == max(s)] * dt
        A_1 = (1 / 2) * s[s == max(s)]
        sigma_1 = (2 / 3) * f[((2 / 3) * s[s == max(s)] - s) == min((2 / 3) * s[s == max(s)] - s)] * dt
        gaus_1 = gaussian(f * dt, A_1, mu_1, sigma_1)

        try:
            mu_2 = f[s == np.sort(s[max_bool(s)])[-2]] * dt
            A_2 = (1 / 2) * s[s == np.sort(s[max_bool(s)])[-2]]
            sigma_2 = (2 / 3) * f[((2 / 3) * s[s == np.sort(s[max_bool(s)])[-2]] - s)
                                  == min((2 / 3) * s[s == np.sort(s[max_bool(s)])[-2]] - s)] * dt
            gaus_2 = gaussian(f * dt, A_2, mu_2, sigma_2)
        except:
            gaus_2 = gaussian(f * dt, A_1, mu_1, sigma_1)


        mu_3 = (mu_1 + mu_2) / 2
        A_3 = np.asarray([(1 / 4) * s[np.abs(f * dt - mu_3) == min(np.abs(f * dt - mu_3))][0]])
        sigma_3 = 4 * np.abs(mu_1 - mu_2)
        gaus_3 = gaussian(f * dt, A_3, mu_3, sigma_3)

        x0 = np.zeros(6)
        x0[0] = A_1
        x0[1] = A_2
        x0[2] = A_3
        x0[3] = sigma_1
        x0[4] = sigma_2
        x0[5] = sigma_3

        thetas = gaus_param(x0, f * dt, mu_1, mu_2, mu_3, s, method=method).x
        f_range = [(mu_1 - 2.5 * thetas[3])[0], (mu_1 + 2.5 * thetas[3])[0]]

        if plot:
            plt.plot(np.asarray(time_series_resid))
            plt.show()
            s, f = mlab.psd(np.asarray(time_series_resid), Fs=1 / dt)
            plt.title('Gaussian Function Initialisation')
            plt.plot(f * dt, s, label='Power-spectral density')
            plt.plot(f * dt, gaus_1, '--', label=r'$g_1(f) = A_1^{(0)}e^{\frac{(f-\mu_1)^2}{\sigma^{(0)2}_1}}$', linewidth=2)
            plt.plot(f * dt, gaus_2, '--', label=r'$g_2(f) = A_2^{(0)}e^{\frac{(f-\mu_2)^2}{\sigma^{(0)2}_2}}$', linewidth=2)
            plt.plot(f * dt, gaus_3, '--', label=r'$g_3(f) = A_3^{(0)}e^{\frac{(f-\mu_3)^2}{\sigma^{(0)2}_3}}$', linewidth=2)
            plt.plot(f * dt, gaus_1 + gaus_2 + gaus_3, '--',
                     label=r'$\sum_{i=1}^{3}g_i(f)$', linewidth=2)
            plt.plot(mu_1 * np.ones(100), np.linspace(np.min(s), 1.1 * np.max(s), 100), '--', label=r'$\mu_1$')
            plt.plot(mu_2 * np.ones(100), np.linspace(np.min(s), 1.1 * np.max(s), 100), '--', label=r'$\mu_2$')
            plt.plot(mu_3 * np.ones(100), np.linspace(np.min(s), 1.1 * np.max(s), 100), '--', label=r'$\mu_3$')
            plt.legend(loc='best', fontsize=8)
            plt.xlabel('Standardised Frequency')
            plt.ylabel('Spectral Density')
            plt.xlim(-0.005, 0.255)
            plt.xticks([0, 0.05, 0.1, 0.15, 0.2, 0.25])
            plt.show()
            plt.title('Gaussian Function Optimised')
            plt.plot(f * dt, s, label='Power-spectral density')
            plt.plot(f * dt, gaussian(f * dt, thetas[0], mu_1[0], thetas[3]),
                     '--', label=r'$g^{opt}_1(f) = A_1^{opt}e^{\frac{(f-\mu_1)^2}{\sigma^{opt2}_1}}$', linewidth=2)
            plt.plot(f * dt, gaussian(f * dt, thetas[1], mu_2[0], thetas[4]),
                     '--', label=r'$g^{opt}_2(f) = A_2^{opt}e^{\frac{(f-\mu_2)^2}{\sigma^{opt2}_2}}$', linewidth=2)
            plt.plot(f * dt, gaussian(f * dt, thetas[2], mu_3[0], thetas[5]),
                     '--', label=r'$g^{opt}_3(f) = A_3^{opt}e^{\frac{(f-\mu_3)^2}{\sigma^{opt2}_3}}$', linewidth=2)
            plt.plot(f * dt, gaussian(f * dt, thetas[0], mu_1[0], thetas[3]) +
                     gaussian(f * dt, thetas[1], mu_2[0], thetas[4]) +
                     gaussian(f * dt, thetas[2], mu_3[0], thetas[5]), '--',
                     label=r'$\sum_{i=1}^{3}g^{opt}_i(f)$', linewidth=2)
            plt.plot(f_range[0] * np.ones(101), np.linspace(0, 1.1 * 2 * A_1, 101), 'k--')
            plt.plot(f_range[1] * np.ones(101), np.linspace(0, 1.1 * 2 * A_1, 101), 'k--', label='Frequency bounds')
            plt.legend(loc='best', fontsize=8)
            plt.xlabel('Standardised Frequency')
            plt.ylabel('Spectral Density')
            plt.xlim(-0.005, 0.255)
            plt.xticks([0, 0.05, 0.1, 0.15, 0.2, 0.25])
            plt.show()
        L = int(1.2 * (1 / dt) * (1 / f[np.max(s) == s]))
        fft_spectrum = sp.fft.fft(time_series_resid)
        fft_freq = sp.fft.fftfreq(len(time_series_resid))
        if f_range[0] < 0:
            fft_spectrum_band = fft_spectrum * (fft_freq < f_range[1])
            fft_spectrum_band = fft_spectrum_band * (fft_freq > -f_range[1])
        else:
            fft_spectrum_band = fft_spectrum * (np.r_[fft_freq < f_range[1]] & np.r_[fft_freq > f_range[0]])
            fft_spectrum_band += fft_spectrum * (np.r_[fft_freq > -f_range[1]] & np.r_[fft_freq < -f_range[0]])
        trend_est_1 = sp.fft.ifft(fft_spectrum_band)
        trend_est_1 = np.real(trend_est_1)

        # decomposition - MODIFIED embedding
        X = np.zeros((len(np.asarray(trend_est_1)), len(np.asarray(trend_est_1)) - L + 1))
        for col in range(len(np.asarray(trend_est_1)) - L + 1):
            X[:, col] = np.hstack((np.asarray(trend_est_1)[col:int(len(np.asarray(trend_est_1) + col))], np.asarray(trend_est_1)[:col]))

        # decomposition - singular value decomposition (eigen-vectors in columns)
        eigen_values, eigen_vectors = np.linalg.eig(np.matmul(X, X.T))
        V_storage = np.real(np.matmul(X.T, eigen_vectors[:, 0].reshape(-1, 1)) / np.sqrt(eigen_values[0]))
        X_estimate = np.real(np.sqrt(eigen_values[0]) * np.matmul(eigen_vectors[:, 0].reshape(-1, 1), V_storage.T))

        # reconstruction - averaging
        trend_est_2 = X_estimate[:, 0].copy()
        averaging_vector = (len(time_series_resid) - L + 1) * np.ones_like(trend_est_2)
        for col in range(1, len(time_series_resid) - L + 1):
            trend_est_2[col:] += X_estimate[:int(len(trend_est_2[col:])), col]
            trend_est_2[:col] += X_estimate[int(len(trend_est_2[col:])):, col]
        trend_est_2 /= averaging_vector
        trend_est_2 = np.real(trend_est_2)

        if plot:
            plt.plot(time_series_resid, label='Detrended time series')
            plt.title('Downsampled and Modified Embedding Estimates')
            plt.plot(trend_est_1, '--', label='Downsampled estimate', linewidth=2)
            plt.plot(trend_est_2, '--', label='Modified embedding estimate', linewidth=2)
            plt.legend(loc='best')
            # plt.xlabel('Months')
            # plt.ylim(-390, 390)
            plt.show()

        # if sum(trend_est_2 * (time_series_resid - trend_est_2)) > 0:
        if sum((time_series_resid - trend_est_2) ** 2) < sum((time_series_resid - trend_est_1) ** 2):
            trend_est = trend_est_2
        else:
            trend_est = trend_est_1

        # scaling factor - top

        a_opt = scaling_factor(residual_time_series=time_series_resid, trend_estimate=trend_est).x
        if plot:
            plt.title('Optimisation of Scaling Factor')
            plt.plot(np.asarray(time_series_resid), label=r'$v_i(t)$')
            plt.plot(trend_est, '--', label=r'$g_i(t)$', linewidth=2)
            plt.plot(a_opt * trend_est, '--', label=r'$\tilde{g}_i(t)$', linewidth=2)
            plt.legend(loc='best')
            # plt.xlabel('Months')
            # plt.ylim(-390, 390)
            plt.show()
        trend_est *= a_opt

        # scaling factor - bottom

        time_series_resid -= trend_est

        try:
            if sum(np.abs(trend_est)) == 0:
                time_series_est_mat = np.vstack((time_series_est_mat, time_series_resid))
                time_series_resid = np.zeros_like(time_series_resid)
            else:
                time_series_est_mat = np.vstack((time_series_est_mat, trend_est))
        except:
            time_series_est_mat = trend_est.reshape(1, -1)

        nmse = sum(time_series_resid ** 2) / sum((time_series_orig - np.mean(time_series_orig)) ** 2)

    time_series_est_mat = np.vstack((time_series_est_mat, time_series_resid))

    return time_series_est_mat




#     ________
#            /
#      \    /
#       \  /
#        \/

# Main reference: Hassani (2007)
# Hassani, H. (2007). Singular Spectrum Analysis: Methodology and Comparison.
# Cardiff University and Central Bank of the Islamic Republic of Iran.
# https://mpra.ub.uni-muenchen.de/4991/

import numpy as np
import seaborn as sns
from scipy.stats import norm
import matplotlib.pyplot as plt

np.random.seed(0)

sns.set(style='darkgrid')


def CovRegpy_ssa(time_series, L, est=3, plot=False, KS_test=False, plot_KS_test=False, KS_scale_limit=1.0,
                 max_eig_ratio=0.0001, KS_start=1, KS_end=3, KS_interval=1):
    """
    Singular Spectrum Analysis (SSA) as in Hassani (2007).

    Parameters
    ----------
    time_series : real ndarray
        Time series on which SSA will be applied.

    L : positive integer
        Embedding dimension as in Hassani (2007).

    est : positive integer
        Number of components to use in trend estimation. Necessarily true that est <= L.

    plot : boolean
        Whether to plot time series and trend estimation before returning trend and decomposition (one plot at end).

    KS_test : boolean
        Whether to test for optimal L and est under Kolmogorov-Smirnov cumulative distribution conditions.

    plot_KS_test : boolean
        Whether to plot all intermediate Kolmogorov-Smirnov tests.

    KS_scale_limit : positive float
        Kolmogorov-Smirnov conditions optimised subject to a minimum standard deviation.
        Without this limit, 'optimised' trend estimate would be the exact time series.
        Set to zero or small for a test - see example.

    max_eig_ratio : float
        Not in original algorithm, but we conjecture that is essential for time-saving and prevention of inclusion
        of probably inconsequential components i.e. components with minimal contribution to variation.

    KS_start : integer
        Beginning value of grid search for Kolmogorov Smirnov Test.

    KS_end : integer
        Possible end value of grid search for Kolmogorov Smirnov Test.
        Only possible as we take the minimum of KS_end and len(time_series) / 3.

    KS_interval : integer
        Interval of grid search for Kolmogorov Smirnov Test.

    Returns
    -------
    time_series_est : real ndarray
        SSA trend estimation of time series using est number of components to estimate trend.

    time_series_decomp : real ndarray
        SSA decomposition of time series into est number of components.

    Notes
    -----
    Kolmogorov-Smirnov Singular Spectrum Analysis (KS-SSA) is experimental and effective.

    """

    if not isinstance(time_series, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('time_series must be of type np.ndarray.')
    if np.array(time_series).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('time_series must only contain floats.')
    if (not isinstance(L, int)) or (not L > 0) or (not L < len(time_series)):
        raise ValueError('\'L\' must be a positive integer of appropriate magnitude: L < len(time_series).')
    if (not isinstance(est, int)) or (not est > 0) or (not est <= L):
        raise ValueError('\'est\' must be a positive integer of appropriate magnitude: est <= L.')
    if not isinstance(plot, bool):
        raise TypeError('\'plot\' must be boolean.')
    if not isinstance(KS_test, bool):
        raise TypeError('\'KS_test\' must be boolean.')
    if not isinstance(plot_KS_test, bool):
        raise TypeError('\'plot_KS_test\' must be boolean.')
    if (not isinstance(KS_scale_limit, float)) or KS_scale_limit <= 0:
        raise ValueError('\'KS_scale_limit\' must be a positive float.')
    if (not isinstance(max_eig_ratio, float)) or max_eig_ratio <= 0 or max_eig_ratio >= 1:
        raise ValueError('\'max_eig_ratio\' must be a float percentage between 0 and 1.')
    if (not isinstance(KS_start, int)) or (not KS_start > 0) or (not KS_start <= len(time_series) / 3):
        raise ValueError('\'KS_start\' must be a positive integer of appropriate magnitude: '
                         'KS_start <= len(time_series) / 3.')
    if (not isinstance(KS_end, int)) or (not KS_end > KS_start) or (not KS_end <= len(time_series) / 3):
        raise ValueError('\'KS_end\' must be a positive integer of appropriate magnitude: '
                         'KS_end > KS_start and KS_end <= len(time_series) / 3.')
    if (not isinstance(KS_interval, int)) or (not KS_interval > 0) or (not KS_interval < (KS_end - KS_start)):
        raise ValueError('\'KS_interval\' must be a positive integer of appropriate magnitude: '
                         'KS_interval < (KS_end - KS_start).')

    KS_value = 1000

    fig_plot_count = 0

    if KS_test:
        for L_test in np.arange(KS_start, int(min(len(time_series) / 3, KS_end)), KS_interval):
            for est_test in np.arange(1, int(L_test)):

                prev_test_value = 1.0
                if int(est_test) > 1:
                    prev_test_value = test_value.copy()

                trend = CovRegpy_ssa(time_series, L=int(L_test), est=int(est_test), KS_test=False)[0]

                errors = time_series - trend
                std_of_errors = np.std(errors)
                x = np.linspace(-5 * std_of_errors, 5 * std_of_errors, len(errors))
                sorted_errors = np.sort(errors)
                ks_vector = np.zeros_like(x)
                for error in sorted_errors:
                    for index in range(len(x)):
                        if x[index] < error:
                            ks_vector[index] += 1
                ks_vector = 1 - ks_vector / len(x)
                test_value = np.max(np.abs(norm.cdf(x, scale=std_of_errors) - ks_vector))
                x_test_value = x[np.max(np.abs(norm.cdf(x, scale=std_of_errors) - ks_vector)) ==
                                 np.abs(norm.cdf(x, scale=std_of_errors) - ks_vector)]
                if plot_KS_test:
                    plt.title('Kolmogorov-Smirnoff Test')
                    plt.plot(x, norm.cdf(x, scale=std_of_errors), label='Cumulative distribution')
                    plt.plot(x, ks_vector, label='Test distribution')
                    plt.plot(x_test_value * np.ones(100),
                             np.linspace(norm.cdf(x_test_value, scale=std_of_errors),
                                         ks_vector[np.max(np.abs(norm.cdf(x, scale=std_of_errors) - ks_vector)) ==
                                                   np.abs(norm.cdf(x, scale=std_of_errors) - ks_vector)], 100),
                             'k', label='KS distance')
                    plt.legend(loc='best')
                    plt.text(x[0], 0.5, f'KS Test Statistic = {np.round(test_value, 5)}', fontsize=8)
                    plt.xticks(fontsize=8)
                    plt.xlabel('Errors', fontsize=10)
                    plt.yticks(fontsize=8)
                    plt.ylabel('Cumulative Distribution', fontsize=10)
                    plt.show()

                if (test_value < KS_value) and (std_of_errors > KS_scale_limit):
                    KS_value = test_value
                    L_opt = int(L_test)
                    est_opt = int(est_test)
                if test_value == prev_test_value:
                    break
                prev_test_value = test_value.copy()

        print(L_opt)
        print(est_opt)

        L = L_opt
        est = est_opt

    # decomposition - embedding
    X = np.zeros((L, len(time_series) - L + 1))
    for col in range(len(time_series) - L + 1):
        X[:, col] = time_series[col:int(L + col)]

    # decomposition - singular value decomposition
    eigen_values, eigen_vectors = np.linalg.eig(np.matmul(X.T, X))
    eigen_values = np.real(eigen_values)

    eigen_vectors = np.real(eigen_vectors)
    # eigen-vectors in columns
    V_storage = {}
    X_storage = {}
    for i in range(sum(eigen_values > max_eig_ratio * np.max(eigen_values))):
        V_storage[i] = np.matmul(X, eigen_vectors[:, i].reshape(-1, 1)) / np.sqrt(eigen_values[i])
        X_storage[i] = np.sqrt(eigen_values[i]) * np.matmul(eigen_vectors[:, i].reshape(-1, 1), V_storage[i].T)

    # reconstruction - grouping
    X_estimate = np.zeros_like(X)
    for j in range(min(len(X_storage), est)):
        X_estimate += X_storage[j].T

    # reconstruction - averaging
    time_series_est = np.zeros_like(time_series)
    averaging_vector = L * np.ones_like(time_series_est)
    for col in range(len(time_series) - L + 1):
        time_series_est[col:int(L + col)] += X_estimate[:, col]
        if col < L:
            averaging_vector[col] -= (L - col - 1)
            averaging_vector[-int(col + 1)] -= (L - col - 1)
    time_series_est /= averaging_vector

    # Decomposing Singular Spectrum Analysis (D-SSA)
    time_series_decomp = np.zeros((est, len(time_series_est)))

    for comp in range(min(len(X_storage), est)):
        comp_est = np.zeros_like(time_series)
        for col in range(len(time_series) - L + 1):
            comp_est[col:int(L + col)] += X_storage[comp][col, :]
        time_series_decomp[comp, :] = comp_est / averaging_vector

    if plot:
        plt.plot(time_series)
        plt.plot(time_series_est, '--')
        plt.show()

    return time_series_est, time_series_decomp




from setuptools import setup

setup(
  name = 'CovRegpy',
  packages = ['CovRegpy'],
  version = '0.0.5',
  license='cc-by-nc-4.0',
  description = 'Regularised covariance regression software based on Hoff and Niu (2012).',
  long_description = 'Regularised covariance regression software based on Hoff and Niu (2012) - see https://arxiv.org/pdf/1102.5721. This package was developed out of research performed by Cole van Jaarsveldt, Gareth W. Peters, Matthew Ames, and Mike Chantler. This package was built entirely using Python 3.11.5 - Python guarantees backwards compatibility which should ensure that this software package functions as expected on all future Python versions. See:\\\\ van Jaarsveldt, C., Peters, G., Ames, M., & Chantler, M. (2024) Package CovRegpy: Regularized covariance regression and forecasting in Python. Annals of Actuarial Science, First View. doi:10.1017/S1748499524000101 url: https://dx.doi.org/10.1017/S1748499524000101',
  author = 'Cole van Jaarsveldt',
  author_email = 'colevj0303@gmail.com',
  url = 'https://github.com/Cole-vJ/CovRegpy.git',
  download_url = 'https://github.com/Cole-vJ/CovRegpy/archive/refs/tags/0.0.5.tar.gz',
  keywords = ['Portfolio Optimisation', 'Regularised Covariance Regression (RCR)', 'Empirical Mode Decomposition (EMD)', 'Singular Spectrum Analysis (SSA)', 'Singular Spectrum Decomposition (SSD)', 'X11', 'Implicit Factors', 'Risk Premia Parity', 'Risk Parity', 'Long\Short Equity'],
  install_requires=[
          'numpy',
          'seaborn',
	  'group-lasso',
	  'AdvEMDpy',
	  'pandas',
	  'matplotlib',
	  'yfinance',
  	  'mdlp-discretization',
	  'arch',
  	  'notebook',
      ],
  classifiers=[
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: Free for non-commercial use',
    'Programming Language :: Python :: 3.11',
  ],
)




#     ________
#            /
#      \    /
#       \  /
#        \/

# Main reference: Shiskin, Young, and Musgrave (1967)
# J. Shiskin, A. Young, and J. Musgrave. 1967.
# The X-11 Variant of the Census Method II Seasonal Adjustment Program.
# Technical Report 15. U.S. Department of Commerce, Washington D.C.

import numpy as np
import seaborn as sns

from scipy.interpolate import interp1d

sns.set(style='darkgrid')


def henderson_kernel(order=13, start=None, end=None):
    """
    Henderson Kernel.

    Parameters
    ----------
    order : odd positive integer
        Order of filter.

    start : negative integer or None
        Used in asymmetric weighting. If None then symmetric weighting.

    end : positive integer or None
        Used in asymmetric weighting. If None then symmetric weighting.

    Returns
    -------
    y : real ndarray
        Weights for Henderson filter.

    Notes
    -----
    Has easily calculable asymmetric weighting.
    Exact Henderson Kernel - differs slightly from classical Henderson smoother.
    Renormalised when incomplete - does nothing when complete as weights sum to one.

    """

    if start is None:
        start = -int((order - 1) / 2)
    if end is None:
        end = int((order - 1) / 2)

    if (not isinstance(order, int)) or (not order > 0) or (not order % 2 == 1):
        raise ValueError('\'order\' must be a positive odd integer.')
    if (not isinstance(start, int)) or (start > 0) or (start < -int((order - 1) / 2)):
        raise ValueError('\'start\' must be non-positive integer of correct magnitude.')
    if (not isinstance(end, int)) or (end < 0) or (end > int((order - 1) / 2)):
        raise ValueError('\'end\' must be non-negative integer of correct magnitude.')

    t = np.asarray(range(start, end + 1)) / (int((order - 1) / 2) + 1)
    y = (15 / 79376) * (5184 - 12289 * t ** 2 + 9506 * t ** 4 - 2401 * t ** 6) * ((2175 / 1274) - (1372 / 265) * t ** 2)
    y = y / sum(y)

    return y


def henderson_weights(order=13, start=None, end=None):
    """
    Classical Henderson weights.

    Parameters
    ----------
    order : odd positive integer
        Order of filter.

    start : negative integer or None
        Used in asymmetric weighting. If None then symmetric weighting.

    end : positive integer or None
        Used in asymmetric weighting. If None then symmetric weighting.

    Returns
    -------
    y : real ndarray
        Weights for classical Henderson weights.

    Notes
    -----
    Does not have easily calculable asymmetric weighting.
    Renormalised when incomplete - does nothing when complete as weights sum to one.

    """
    if start is None:
        start = -int((order - 1) / 2)
    if end is None:
        end = int((order - 1) / 2)

    if (not isinstance(order, int)) or (not order > 0) or (not order % 2 == 1):
        raise ValueError('\'order\' must be a positive odd integer.')
    if (not isinstance(start, int)) or (start > 0) or (start < -int((order - 1) / 2)):
        raise ValueError('\'start\' must be non-positive integer of correct magnitude.')
    if (not isinstance(end, int)) or (end < 0) or (end > int((order - 1) / 2)):
        raise ValueError('\'end\' must be non-negative integer of correct magnitude.')
    if start != -end:
        raise ValueError('\'start\' and \'end\' must be equal and opposite.')

    p = int((order - 1) / 2)
    n = p + 2
    vector = np.asarray(range(start, (end + 1)))
    y = (315 * ((n - 1) ** 2 - vector ** 2) * (n ** 2 - vector ** 2) * ((n + 1) ** 2 - vector ** 2) *
         (3 * n ** 2 - 16 - 11 * vector ** 2)) / \
        (8 * n * (n ** 2 - 1) * (4 * n ** 2 - 1) * (4 * n ** 2 - 9) * (4 * n ** 2 - 25))
    y = y / sum(y)

    return y


def henderson_ma(time_series, order=13, method='kernel'):
    """
    Henderson filter.

    Parameters
    ----------
    time_series : real ndarray
        Time series to be Henderson filtered.

    order : odd positive integer
        Order of Henderson filter.

    method : string
        Technique to be used in calculation of weights.

    Returns
    -------
    henderson_filtered_time_series : real ndarray
        Henderson filtered time series.

    Notes
    -----
    Only two options. Classical weights and kernel weights.

    Require asymmetric weights that sum to approximately one on the edges - multiple options:
        (1) use asymmetric filter (truncate and renormalise) - 'renormalise'.
        (2) extrapolate and use symmetric filter (X11ARIMA) - not yet added.
        (3) Reproducing Kernel Hilbert Space Method - 'kernel'.
        (4) Classical asymmetric results (unknown calculation) - not yet added.

    """
    if not isinstance(time_series, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('time_series must be of type np.ndarray.')
    if np.array(time_series).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('time_series must only contain floats.')
    if (not isinstance(order, int)) or (not order > 0) or (not order % 2 == 1):
        raise ValueError('\'order\' must be a positive odd integer.')
    if method not in {'renormalise', 'kernel'}:
        raise ValueError('\'method\' not an acceptable value.')

    henderson_filtered_time_series = np.zeros_like(time_series)

    if method == 'renormalise':
        weights = henderson_weights(order=order)
    elif method == 'kernel':
        weights = henderson_kernel(order=order)

    for k in range(len(time_series)):
        if k < ((order - 1) / 2):
            if method == 'renormalise':
                asymmetric_weights = henderson_weights(order=order, start=(0 - k))
            elif method == 'kernel':
                asymmetric_weights = henderson_kernel(order=order, start=(0 - k))
            henderson_filtered_time_series[k] = \
                np.sum(asymmetric_weights * time_series[:int(k + ((order - 1) / 2) + 1)])
        elif k > len(time_series) - ((order - 1) / 2) - 1:
            if method == 'renormalise':
                asymmetric_weights = henderson_weights(order=order, end=(len(time_series) - k - 1))
            elif method == 'kernel':
                asymmetric_weights = henderson_kernel(order=order, end=(len(time_series) - k - 1))
            henderson_filtered_time_series[k] = \
                np.sum(asymmetric_weights * time_series[int(k - ((order - 1) / 2)):])
        else:
            henderson_filtered_time_series[k] = \
                np.sum(weights * time_series[int(k - ((order - 1) / 2)):int(k + ((order - 1) / 2) + 1)])

    return henderson_filtered_time_series


def seasonal_ma(time_series, factors='3x3', seasonality='annual'):
    """
    Seasonal moving-average filter.

    Parameters
    ----------
    time_series : real ndarray
        Time series to be seasonally filtered.

    ########################################################
    # Must automate factors creation in subsequent version #
    ########################################################

    factors : string
        Seasonal filter to be applied.

    seasonality : string
        Seasonality to be used - either annual or quarterly.

    Returns
    -------
    seasonal_filtered_time_series : real ndarray
        Seasonally filtered time series.

    Notes
    -----
    Need to make an automated factor-weighting calculation.

    """

    if not isinstance(time_series, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('time_series must be of type np.ndarray.')
    if np.array(time_series).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('time_series must only contain floats.')
    if factors not in {'3x3', '3x5', '3x7', '3x9'}:
        raise ValueError('\'factors\' not an acceptable value.')
    if seasonality not in {'annual', 'quarterly'}:
        raise ValueError('\'seasonality\' not an acceptable value.')

    seasonal_filtered_time_series = np.zeros_like(time_series)

    if factors == '3x3':
        weighting = np.asarray((1 / 9, 2 / 9, 1 / 3, 2 / 9, 1 / 9))
        season_window_width = 5
    elif factors == '3x5':
        weighting = np.asarray((1 / 15, 2 / 15, 1 / 5, 1 / 5, 1 / 5, 2 / 15, 1 / 15))
        season_window_width = 7
    elif factors == '3x7':
        weighting = np.asarray((1 / 21, 2 / 21, 1 / 7, 1 / 7, 1 / 7, 1 / 7, 1 / 7, 2 / 21, 1 / 21))
        season_window_width = 9
    elif factors == '3x9':
        weighting = np.asarray((1 / 27, 2 / 27, 1 / 9, 1 / 9, 1 / 9, 1 / 9, 1 / 9, 1 / 9, 1 / 9, 2 / 27, 1 / 27))
        season_window_width = 11

    if seasonality == 'annual':
        for month in range(12):

            month_bool = (np.asarray(range(len(time_series))) % 12 - month == 0)
            index_values = np.asarray(range(len(time_series)))[month_bool]

            for point in np.asarray(range(len(index_values)))[
                         int((season_window_width - 1) / 2):int(len(index_values) - ((season_window_width - 1) / 2))]:
                relative_points = \
                    time_series[index_values[int(point - ((season_window_width - 1) / 2)):int(
                        point + ((season_window_width + 1) / 2))]]

                seasonal_filtered_time_series[index_values[point]] = \
                    sum(weighting * relative_points)

        left_edge = seasonal_filtered_time_series[int(((season_window_width - 1) / 2) * 12):int(((season_window_width + 1) / 2) * 12)]
        for left_index in range(1, int((season_window_width - 1) / 2)):
            left_edge = \
                np.append(left_edge,
                          seasonal_filtered_time_series[int(((season_window_width - 1) / 2) * 12):int(((season_window_width + 1) / 2) * 12)])

        seasonal_filtered_time_series[:int(((season_window_width - 1) / 2) * 12)] = left_edge

        right_edge = seasonal_filtered_time_series[int(len(seasonal_filtered_time_series) - ((season_window_width + 1) / 2) * 12):int(len(seasonal_filtered_time_series) - ((season_window_width - 1) / 2) * 12)]
        for right_index in range(1, int((season_window_width - 1) / 2)):
            right_edge = np.append(right_edge, seasonal_filtered_time_series[int(len(seasonal_filtered_time_series) - ((season_window_width + 1) / 2) * 12):int(len(seasonal_filtered_time_series) - ((season_window_width - 1) / 2) * 12)])

        seasonal_filtered_time_series[int(len(seasonal_filtered_time_series) - ((season_window_width - 1) / 2) * 12):] = right_edge

    elif seasonality == 'quarterly':
        for quarter in range(4):

            month_bool = (np.asarray(range(len(time_series))) % 4 - quarter == 0)
            index_values = np.asarray(range(len(time_series)))[month_bool]

            for point in np.asarray(range(len(index_values)))[
                         int((season_window_width - 1) / 2):int(len(index_values) - ((season_window_width - 1) / 2))]:
                relative_points = \
                    time_series[index_values[int(point - ((season_window_width - 1) / 2)):int(
                        point + ((season_window_width + 1) / 2))]]

                seasonal_filtered_time_series[index_values[point]] = \
                    sum(weighting * relative_points)

        seasonal_filtered_time_series[:int(((season_window_width - 1) / 2) * 4)] = np.append(
            seasonal_filtered_time_series[
            int(((season_window_width - 1) / 2) * 4):int(((season_window_width + 1) / 2) * 4)],
            seasonal_filtered_time_series[
            int(((season_window_width - 1) / 2) * 4):int(((season_window_width + 1) / 2) * 4)])

        seasonal_filtered_time_series[
        int(len(seasonal_filtered_time_series) - ((season_window_width - 1) / 2) * 4):] = np.append(
            seasonal_filtered_time_series[
            int(len(seasonal_filtered_time_series) - ((season_window_width + 1) / 2) * 4):int(
                len(seasonal_filtered_time_series) - ((season_window_width - 1) / 2) * 4)],
            seasonal_filtered_time_series[int(
                len(seasonal_filtered_time_series) - ((season_window_width + 1) / 2) * 4):int(
                len(seasonal_filtered_time_series) - ((season_window_width - 1) / 2) * 4)])

    return seasonal_filtered_time_series


def CovRegpy_X11(time, time_series, seasonality='annual', seasonal_factor='3x3',
                 trend_window_width_1=13, trend_window_width_2=13, trend_window_width_3=13):
    """
    Standard X11 decomposition method.

    Parameters
    ----------
    time : real ndarray
        Time over which time series is defined.

    time_series : real ndarray
        Time series to be filtered.

    seasonality : string
        Seasonality to be used - either annual or quarterly.

    seasonal_factor : string
        Seasonal weights to be used.

    trend_window_width_1 : odd positive integer
        Trend window width to be used on first iteration.

    trend_window_width_2 : odd positive integer
        Trend window width to be used on second iteration.

    trend_window_width_3 : odd positive integer
        Trend window width to be used on third iteration.

    Returns
    -------
    final_estimate_trend : real ndarray
        Trend estimate of X11 filtered method.

    final_estimate_season : real ndarray
        Seasonal estimate of X11 filtered method.

    final_estimate_irregular : real ndarray
        Irregular estimate (error terms) of X11 filtered method.

    Notes
    -----
    Standard X11 method.

    """

    if not isinstance(time, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('time must be of type np.ndarray.')
    if np.array(time).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('time must only contain floats.')
    if not isinstance(time_series, (type(np.asarray([[1.0, 2.0], [3.0, 4.0]])))):
        raise TypeError('time_series must be of type np.ndarray.')
    if np.array(time_series).dtype != np.array([[1., 1., 1., 1.]]).dtype:
        raise TypeError('time_series must only contain floats.')
    if len(time) != len(time_series):
        raise ValueError('time_series and time are incompatible lengths.')
    if seasonality not in {'annual', 'quarterly'}:
        raise ValueError('\'seasonality\' not an acceptable value.')
    if seasonal_factor not in {'3x3', '3x5', '3x7', '3x9'}:
        raise ValueError('\'seasonal_factor\' not an acceptable value.')
    if (not isinstance(trend_window_width_1, int)) or (not trend_window_width_1 > 0) or (not trend_window_width_1 % 2 == 1):
        raise ValueError('\'trend_window_width_1\' must be a positive odd integer.')
    if (not isinstance(trend_window_width_2, int)) or (not trend_window_width_2 > 0) or (not trend_window_width_2 % 2 == 1):
        raise ValueError('\'trend_window_width_2\' must be a positive odd integer.')
    if (not isinstance(trend_window_width_3, int)) or (not trend_window_width_3 > 0) or (not trend_window_width_3 % 2 == 1):
        raise ValueError('\'trend_window_width_3\' must be a positive odd integer.')

    # step 1
    # initial estimate of trend-cycle
    first_estimate_trend = np.zeros_like(time_series)

    for point in \
            range(len(time_series))[int((trend_window_width_1 - 1) / 2):int(len(time_series) -
                                                                            ((trend_window_width_1 - 1) / 2))]:

        first_estimate_trend[point] = \
            np.mean(time_series[int(point -
                                    ((trend_window_width_1 - 1) / 2)):int(point +
                                                                          ((trend_window_width_1 - 1) / 2))])

    # interpolate edges
    # relevant to remove some trend at edges
    interpolation = \
        interp1d(time[int((trend_window_width_1 - 1) / 2): int(len(time_series) -
                                                               ((trend_window_width_1 - 1) / 2))],
                 first_estimate_trend[int((trend_window_width_1 - 1) / 2): int(len(time_series) -
                                                                               ((trend_window_width_1 - 1) / 2))],
                 fill_value='extrapolate')

    left_extrapolate = interpolation(time[:int((trend_window_width_1 - 1) / 2)])
    right_extrapolate = interpolation(time[int(len(time_series) - ((trend_window_width_1 - 1) / 2)):])

    first_estimate_trend[:int((trend_window_width_1 - 1) / 2)] = left_extrapolate
    first_estimate_trend[int(len(time_series) - ((trend_window_width_1 - 1) / 2)):] = right_extrapolate

    # time series without trend-cycle component
    no_trend_cycle = time_series - first_estimate_trend

    # step 2
    # initial estimate of seasonality
    # assume annual for now
    first_estimate_season = seasonal_ma(no_trend_cycle, factors=seasonal_factor, seasonality=seasonality)

    # step 3
    # estimate seasonality adjusted data
    # time series without seasonality component (with trend-cycle component)
    no_seasonality = time_series - first_estimate_season

    # step 4
    # better estimate of trend
    next_estimate_trend = henderson_ma(no_seasonality, trend_window_width_2)
    no_trend_cycle_2 = time_series - next_estimate_trend

    # step 5
    final_estimate_season = seasonal_ma(no_trend_cycle_2, factors=seasonal_factor, seasonality=seasonality)

    # step 6
    no_seasonality_2 = time_series - final_estimate_season

    # step 7
    final_estimate_trend = henderson_ma(no_seasonality_2, trend_window_width_3)

    # step 8
    final_estimate_irregular = no_seasonality_2 - final_estimate_trend

    return final_estimate_trend, final_estimate_season, final_estimate_irregular
